import json

import pandas as pd
from dateutil.relativedelta import relativedelta
from empyrical import sortino_ratio
from framework import filter_weekend, dict_remove, get_config, component, autowired
from api import FundOptimize, Navs, BusinessException, Datum
from fund_pool.dao import robo_optimize_pool as rop


@component
class SortinoFundOptimize(FundOptimize):

    _navs: Navs = None
    _datum: Datum = None

    @autowired
    def __init__(self, navs: Navs = None, datum: Datum = None):
        self._navs = navs
        self._datum = datum
        optimize_config = get_config(__name__)
        self._config = [{
            **x,
            'name': [f"sortino_{y[1]}_{y[0]}" for y in x.items() if y[0] != 'weight'][0]
        } for x in optimize_config['sortino_weight']] if 'sortino_weight' in optimize_config else []

    def find_optimize(self, fund_ids, day):
        if not self._config:
            raise BusinessException(f"find optimize, but not found sortino config.")
        start = filter_weekend(sorted([day - relativedelta(days=1, **dict_remove(x, ('weight', 'name'))) for x in self._config])[0])
        fund_navs = pd.DataFrame(self._navs.get_fund_navs(fund_ids=tuple(fund_ids), min_date=start, max_date=day))
        fund_navs.sort_values('nav_date', inplace=True)
        fund_navs = fund_navs.pivot_table(index='nav_date', columns='fund_id', values='nav_cal')
        fund_navs.fillna(method='ffill', inplace=True)
        day_income = round(fund_navs.pct_change(), 4)
        sortino = pd.DataFrame()
        for item in self._config:
            delta_kwargs = item.copy()
            del delta_kwargs['weight'], delta_kwargs['name']
            ratio = dict(sortino_ratio(day_income.truncate(before=(day - relativedelta(**delta_kwargs)))))
            sortino = pd.concat([sortino, pd.DataFrame([ratio], index=[item['name']])])
        sortino = sortino.T
        sortino['score'] = sortino.apply(lambda r: sum([x['weight'] * r[x['name']] for x in self._config]), axis=1)
        sortino.sort_values('score', ascending=False, inplace=True)
        return day_income.columns[sortino.index[0]]

    def get_optimize_pool(self, day):
        last_one = rop.get_last_one(day)
        if not last_one:
            funds = pd.DataFrame(self._datum.get_fund_datums())
            pool = []
            for (category, asset_type), fund_group in funds.groupby(by=['category', 'assetType']):
                if len(fund_group) > 1:
                    pool.append(self.find_optimize(tuple(fund_group['id']), day))
                else:
                    pool.append(fund_group.iloc[0].id)
            rop.insert(day, sorted(pool))
            last_one = rop.get_last_one(day)
        return json.loads(last_one['fund_ids'])


